﻿//---------------------------------------------------------------------------
//
//   ObjectListReorder
//
//          Copyright(C) 2013, Tark
//
//  オブジェクトパネルのオブジェクトリスト操作
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <memory>
#include <sstream>

#include "ObjectListReorderWindow.h"
#include "ObjectListReorder.h"
#include <MQSetting.h>

static ObjectListReorder plugin;

// Constructor
// コンストラクタ
ObjectListReorder::ObjectListReorder()
{
	m_Window = NULL;
	m_bActivate = false;
	setlocale(LC_CTYPE, "");
}

// Destructor
// デストラクタ
ObjectListReorder::~ObjectListReorder()
{
	if(m_Window != NULL){
		delete m_Window;
		m_Window = NULL;
	}
}

//---------------------------------------------------------------------------
//  GetPlugInID
//---------------------------------------------------------------------------
void ObjectListReorder::GetPlugInID(DWORD *Product, DWORD *ID)
{
	*Product = 0xE3F7A830;
	*ID      = 0xBD9A1869;
}

//---------------------------------------------------------------------------
//  GetPlugInName
//---------------------------------------------------------------------------
const char *ObjectListReorder::GetPlugInName(void)
{
	return "ObjectListReorder v.1.2      Copyright(C) 2013-2014, Tark.";
}

//---------------------------------------------------------------------------
//  EnumString
//---------------------------------------------------------------------------
const char *ObjectListReorder::EnumString(void)
{
	return "Reorder";
}


//---------------------------------------------------------------------------
//  Initialize
//---------------------------------------------------------------------------
BOOL ObjectListReorder::Initialize()
{
	int posX = -1, posY = -1;
	{
		MQSetting *setting = OpenSetting();
		setting->Load("ObjectListReorder_activate", m_bActivate, false);
		setting->Load("PosX", posX, -1);
		setting->Load("PosY", posY, -1);
		CloseSetting(setting);
	}
	if (m_Window == NULL){
		MQWindowBase mainwnd = MQWindow::GetMainWindow();
		m_Window = new ObjectListReorderWindow(this, mainwnd);
		if(posX >= 0 && posY >= 0) {
			m_Window->SetPosX(posX);
			m_Window->SetPosY(posY);
		}
	}
	Activate(nullptr, m_bActivate);
	return TRUE;
}

//---------------------------------------------------------------------------
//  Exit
//    アプリケーションの終了
//---------------------------------------------------------------------------
void ObjectListReorder::Exit()
{
	{
		MQSetting *setting = OpenSetting();
		setting->Save("ObjectListReorder_activate", m_bActivate);
		setting->Save("PosX", m_Window->GetPosX());
		setting->Save("PosY", m_Window->GetPosY());
		CloseSetting(setting);
	}
	if (m_Window != NULL){
		delete m_Window;
		m_Window = NULL;
	}
}

//---------------------------------------------------------------------------
//  Activate
//---------------------------------------------------------------------------
BOOL ObjectListReorder::Activate(MQDocument doc, BOOL flag)
{
	m_doc = doc;
	if (m_Window == NULL) {
		return FALSE;
	}
	m_bActivate = flag ? true : false;
	m_Window->SetVisible(m_bActivate);

	return m_bActivate;
}

//---------------------------------------------------------------------------
//  IsActivated
//---------------------------------------------------------------------------
BOOL ObjectListReorder::IsActivated(MQDocument doc)
{
	if (m_Window == NULL) {
		return FALSE;
	}
	return m_bActivate;
}

//---------------------------------------------------------------------------
//  ExecuteCallback
///---------------------------------------------------------------------------
bool ObjectListReorder::ExecuteCallback(MQDocument doc, void *option)
{
	CallbackInfo *info = (CallbackInfo*)option;
	return ((*this).*info->proc)(doc);
}

// コールバックの呼び出し
void ObjectListReorder::Execute(ExecuteCallbackProc proc)
{
	CallbackInfo info;
	info.proc = proc;
	BeginCallback(&info);
}

// プラグインのベースクラスを返す
MQBasePlugin *GetPluginClass()
{
	return &plugin;
}

int ObjectListReorder::moveobj(MQDocument doc, MQObject &target, MQObject to)
{
	if(!target || target == to) {
		return -1;
	}
	MQObject newobj = target->Clone();
	doc->DeleteObject(doc->GetObjectIndex(target));
	int newindex = doc->InsertObject(newobj, to);
	target = newobj;
	return newindex;
}

// 空きスロットを飛ばして次を取得
MQObject ObjectListReorder::next(MQDocument doc, int index, bool forward)
{
	if(index < 0 || index >= doc->GetObjectCount()) {
		return nullptr;
	}
	MQObject target = doc->GetObject(index);
	while(!target || target->GetUniqueID() == 0) {
		if(forward) index++; else  index--;
		if(index < 0 || index >= doc->GetObjectCount()) {
			return nullptr;
		}
		target = doc->GetObject(index);
	}
	return target;
}

// Foldingされている子か
bool ObjectListReorder::isFolding(MQDocument doc, MQObject obj)
{
	MQObject parent = obj;
	while(parent = doc->GetParentObject(parent)) {
		if(parent->GetFolding()) {
			return true;
		}
	}
	return false;
}

MQObject ObjectListReorder::search_next(MQDocument doc, MQObject target, bool forward, bool skipfolding)
{
	int index;
	if(target) {
		index = doc->GetObjectIndex(target);
		if(forward) index++; else  index--;
	}
	else {
		index = forward ? 0 : doc->GetObjectCount() - 1;
	}
	MQObject next(next(doc, index, forward));

	// 親でFoldingのものがあればさらに次
	if(skipfolding) {
		while(isFolding(doc, next)) {
			next = search_next(doc, next, forward, skipfolding);
		}
	}
	return next;
}

std::vector<std::pair<MQObject, int>> ObjectListReorder::getChildList(MQDocument doc, MQObject obj)
{
	std::vector<std::pair<MQObject, int>> list;
	int ocount = doc->GetObjectCount();
	int parentidx = doc->GetObjectIndex(obj);
	int depth = obj->GetDepth();
	for(int i = parentidx + 1; i < ocount; i++) {
		MQObject cobj = doc->GetObject(i);
		if(!cobj) {
			continue;
		}
		if(cobj->GetDepth() <= depth) {
			break;
		}
		list.push_back(std::make_pair(cobj, cobj->GetDepth()));
	}
	return list;
}

int ObjectListReorder::movechildren(MQDocument doc, std::vector<std::pair<MQObject, int>> childlist, MQObject target, int depthdiff)
{
	// 子移動
	int lastindex = 0;
	for(std::pair<MQObject, int> p: childlist) {
		MQObject child = p.first;
		int depth = p.second;
		lastindex = moveobj(doc, child, target);
		child->SetDepth(depth + depthdiff);
		target = child;
	}
	return lastindex;
}

BOOL ObjectListReorder::insert(MQDocument doc)
{
	int curidx = doc->GetCurrentObjectIndex();
	if(curidx < 0) {return -1;}
	MQObject current = doc->GetObject(curidx);
	MQObject to = current;
	if(doc->GetChildObjectCount(current) > 0 &&  current->GetFolding()) {
		MQObject to2 = search_next(doc, current, true, true);
		to = search_next(doc, to2, false, false);
	}

	MQObject target = search_next(doc, nullptr, false, true);
	if(!target) {return -1;}
	if(target == current) {return curidx;}
	int depth = target->GetDepth();
	// targetの子一覧とそのdepthを取得
	std::vector<std::pair<MQObject, int>> childlist = getChildList(doc, target);
	// インサートしたオブジェクトに設定するためのカレントの一つ下オブジェクトのdepth値
	MQObject below = search_next(doc, to, true, true);
	int newdepth = below ? below->GetDepth() : 0;
	// target移動
	int idx = moveobj(doc, target, to);
	target->SetDepth(newdepth);
	// 子に反映させるため、元位置と移動先とのdepthの差分を取得
	int depthdiff = newdepth - depth;
	movechildren(doc, childlist, target, depthdiff);
	return TRUE;
}

BOOL ObjectListReorder::up(MQDocument doc)
{
	// カレントオブジェクトとdepthを取得
	int curidx = doc->GetCurrentObjectIndex();
	if(curidx < 0) {return -1;}
	MQObject current = doc->GetObject(curidx);
	int depth = current->GetDepth();
	// カレントの子一覧とそのdepthを取得
	std::vector<std::pair<MQObject, int>> childlist = getChildList(doc, current);
	// 移動先検索（二つ上）
	MQObject to1 = search_next(doc, current, false, true);
	MQObject to2 = search_next(doc, to1, false, false);
	// カレント移動
	int newidx = moveobj(doc, current, to2);
	if(newidx >= 0) {
		doc->SetCurrentObjectIndex(newidx);
	}
	// 移動先で設定するdepth値は移動先の一つ下オブジェクトと同じにする
	// search_nextで辿られないようにするため一旦currentのFoldingフラグをクリアしておく
	bool currentFolding = (current->GetFolding() == TRUE);
	current->SetFolding(false);
	MQObject below = search_next(doc, current, true, true);
	current->SetFolding(currentFolding);
	int newdepth = below ? below->GetDepth() : 0;
	current->SetDepth(newdepth);
	// 子に反映させるため、元位置と移動先とのdepthの差分を取得
	int depthdiff = newdepth - depth;
	// 子移動
	movechildren(doc, childlist, current, depthdiff);
	return TRUE;
}

BOOL ObjectListReorder::down(MQDocument doc)
{
	// カレントオブジェクトとdepthを取得
	int curidx = doc->GetCurrentObjectIndex();
	if(curidx < 0) {return -1;}
	MQObject current = doc->GetObject(curidx);
	int depth = current->GetDepth();
	// カレントの子一覧とそのdepthを取得
	std::vector<std::pair<MQObject, int>> childlist = getChildList(doc, current);
	// 移動先検索
	MQObject to;
	if(!childlist.empty()) {
		to = search_next(doc, childlist.back().first, true, true);
	}
	else {
		to = search_next(doc, current, true, true);
	}
	MQObject to2 = search_next(doc, to, true, true);
	MQObject to3 = search_next(doc, to2, false, false);
	// カレント移動
	int newidx = moveobj(doc, current, to3);
	if(newidx >= 0) {
		doc->SetCurrentObjectIndex(newidx);
	}
	// 移動先で設定するdepth値は移動先の一つ下オブジェクトと同じにする
	// search_nextで辿られないようにするため一旦currentのFoldingフラグをクリアしておく
	bool currentFolding = (current->GetFolding() == TRUE);
	current->SetFolding(false);
	MQObject below = search_next(doc, current, true, true);
	current->SetFolding(currentFolding);
	int newdepth = below ? below->GetDepth() : 0;
	current->SetDepth(newdepth);
	// 子に反映させるため、元位置と移動先とのdepthの差分を取得
	int depthdiff = newdepth - depth;
	// 子移動
	movechildren(doc, childlist, current, depthdiff);
	return TRUE;
}

BOOL ObjectListReorder::top(MQDocument doc)
{
	int ocount = doc->GetObjectCount();
	MQObject to = nullptr;
	for(int oidx = 0; oidx < ocount; oidx++) {
		// ループ中でオブジェクトを前に移動させると以降のindexがずれていって、前のオブジェクトを再度見ることになるが、
		// 移動させたオブジェクトは選択解除されていて処理対象にならないため、気にしない
		MQObject obj = doc->GetObject(oidx);
		if(!obj || obj->GetUniqueID() == 0) {
			continue;
		}
		if(!obj->GetSelected()) {
			continue;
		}
		int basedepth = -obj->GetDepth();
		// カレントの子一覧とそのdepthを取得
		std::vector<std::pair<MQObject, int>> childlist = getChildList(doc, obj);
		int newidx = moveobj(doc, obj, to);
		to = doc->GetObject(newidx);
		// 子移動
		int childlast = movechildren(doc, childlist, obj, basedepth);
		if(childlast > 0) {
			to = doc->GetObject(childlast);
		}
	}
	return TRUE;
}

BOOL ObjectListReorder::bottom(MQDocument doc)
{
	int ocount = doc->GetObjectCount();
	for(int oidx = 0; oidx < ocount; oidx++) {
		MQObject obj = doc->GetObject(oidx);
		if(!obj || obj->GetUniqueID() == 0) {
			continue;
		}
		if(!obj->GetSelected()) {
			continue;
		}
		int basedepth = -obj->GetDepth();
		// カレントの子一覧とそのdepthを取得
		std::vector<std::pair<MQObject, int>> childlist = getChildList(doc, obj);
		MQObject to = search_next(doc, nullptr, false, false);
		int newidx = moveobj(doc, obj, to);
		// 子移動
		movechildren(doc, childlist, obj, basedepth);
	}
	return TRUE;
}

BOOL ObjectListReorder::gather(MQDocument doc)
{
	int curidx = doc->GetCurrentObjectIndex();
	int ocount = doc->GetObjectCount();
	MQObject target = nullptr;
	std::vector<MQObject> objectlist;
	// 選択オブジェクトを収集対象としてobjectlistに挿入していく
	int depth = INT_MAX;
	for(int oidx = 0; oidx < ocount; oidx++) {
		MQObject obj = doc->GetObject(oidx);
		if(!obj || obj->GetUniqueID() == 0) {
			continue;
		}
		if(obj->GetDepth() > depth) {
			continue; // 収集した選択オブジェクトの子は選択リストに入れないためスキップする
		}
		else {
			depth = INT_MAX;
		}
		if(!obj->GetSelected()) {
			continue;
		}
		if(!target && oidx == curidx) {
			// カレントオブジェクトをtargetにする
			// 収集対象オブジェクトの子であった場合はここに来ないので、移動処理しない
			target = obj;
		}
		else {
			objectlist.push_back(obj);
			depth = obj->GetDepth();
		}
	}
	if(!target || objectlist.empty()) {
		return FALSE;
	}
	// 選択オブジェクトをtargetの子へ移動する
	// 順番を維持するために逆順で処理する
	for(auto it = objectlist.rbegin(); it != objectlist.rend(); ++it) {
		std::vector<std::pair<MQObject, int>> childlist = getChildList(doc, *it);
		int depthdiff = target->GetDepth() + 1 - (*it)->GetDepth();
		moveobj(doc, *it, target);
		// targetの子とする
		(*it)->SetDepth(target->GetDepth() + 1);
		// 子移動
		movechildren(doc, childlist, (*it), depthdiff);
	}
	return TRUE;
}
